/* 
|----------------------------------------------------------------------------|
|  HNNTR.C:  Hebbian Neural Network Trainer    V1.6.d                        |
|                                                                            |                                                                    |
|  By Cristina & Fernando de Castro                                          |
|                                                                            |
|  V1.6 (October 2001) for DJGPP compiler                                    |
|  V1.6.d (February 2013) adapted for DevC++ compiler with conio2 library    |
|----------------------------------------------------------------------------|

Note: The sub-space decomposition of a vector set by the method known as
      Karhunen-Love Transform (principal components analysis - PCA),
      require the solution of the eigenvalues & eigenvectors equation
      for the vector set covariance matrix.

      A Hebbian Neural Network performs the sub-space decomposition
      of a vector set without the need to assemble the vector set
      covariance matrix or to solve any eigenvalue/eigenvector equation .


		|---------------------------------------|
		|                                       |
		|   This program should be compiled     |
		|    with DevC++ compiler V 5.3.0.4     | 
		|      using conio2 library             |
		|                                       |
        |---------------------------------------|
*/

/***************************************************************************
*
* HEADERS:
*
****************************************************************************/

#include<conio2.h> 
#include<stdio.h>
#include<stdlib.h>
#include<process.h>
#include<math.h>
#include<float.h>
#include<ctype.h>
#include<string.h>
#include<time.h>
#include<dos.h>
#include<fcntl.h>
#include<share.h>


/***************************************************************************
*
* PROGRAM DEFINITIONS:
*
****************************************************************************/

#define BUFSIZE 0xFFFF /* file buffer size: 0xFFFF = 64Kb */
#define SMALLBUFSIZE  0x03FF /* file small buffer size 0x03FF = 1Kb  */

#define MAXLINESIZE 65536 /* max line size for  GetSheetNumLines() and NumberOfFloatsInLine() */
#define MAXARGLEN 256   /* max length of an alphanumeric arg */

#define NAVGITV 3   /* squared Euclidean Norm Change averaging interval */
#define PNCC  1 /* period (in epochs) of the squared Euclidean norm convergence check */

/***************************************************************************
*
* COMPILING DIRECTIVES:
*
****************************************************************************/
#pragma GCC diagnostic ignored "-Wwrite-strings"

//#define SHOWEIGENVEC /* if defined, ShowNet() additionally shows the eigenvector set */

//#define DEBUG_FILE "dbg.txt" /* if defined open ascii debug file */

/***************************************************************************
*
* CONSTANTS:
*
****************************************************************************/

#define MAXFLOAT 3.4E38 /* 32 bits DJGPP max float */
#define BEL 7  /* ascii BEL */

#define NALPHALINES 3 /* desired number of alphanumeric lines in parameter file */
#define NFLOATLINES 7 /* desired number of float lines in parameter file */


/***************************************************************************
*
* DATA TYPE STRUCTURES:
*
****************************************************************************/

typedef enum {

  NO_EVENT,           /*  " ", */
  MEMORY_ERR,         /*  "Not enough memory", */
  READ_OPEN_ERR,      /*  "I can't find the file", */
  READ_ERR,           /*  "Error reading file", */
  WRITE_OPEN_ERR,     /*  "I can't open the file", */
  WRITE_ERR,          /*  "I can't write the file", */
  BUFFER_ERR,         /*  "I can't attach buffer to file", */
  SIZE_ERR,           /*  "Specified size doesn't match data size in file", */
  MAXIT_ERR,          /*  "Exceeded max number of iterations in function", */
  LOGIC_ERR,          /*  "Logic error in function", */
  READ_NNET_ERR,      /*  "Incompatible neural network in file", */
  MTXDIM_ERR,         /*  "Invalid matrix dimensions in file", */
  PARF_ERR,           /*  "Wrong parameter file", */
  GFTD_ERR,           /*  "I can't get time & date info from file", */
  MAXFLOAT_READ_ERR,  /*  "Exceeded float maximum absolute value reading file" */
  /* ---*/
  NET_SIZE_ERR       /*  "Read incompatible neural network from file" */
}
EVENT;



typedef struct ParamFileReader{

char *ParameterFile; /* parameter file */
unsigned *FloatLines; /* the array of float lines indexes */
unsigned *AlphaLines; /* the array of alpha lines indexes */
unsigned NumFloatLines; /* the number of float lines */
unsigned NumAlphaLines; /* the number of alpha lines */
char **Sheet; /* parameters file sheet matrix */
char *BufferI; /* input buffer */
FILE *IFp;     /* input filepointer */
unsigned long NumLines;  /* the number of lines read */
unsigned LineMaxLength; /* Sheet max line length */
float **FloatDat; /* float data storage sheet */
char  **AlphaDat;  /* alpha data storage sheet */
unsigned *VectSize; /* the array that contains the size of each row of FloatDat */
char AllocFlag; /* Memory allocation flag: If Allocflag!=0, memory is
                  allocated upon calling Automatic_ParamFileReader() */

}PARFR;



typedef struct Neuron {


double *W; /* synaptic weights vector (size=NSn) */
double y; /* instant neuron output */
float *yc; /* array of converged y values for each TSet sample vector (array size=NSp) */
/* NOTE: yc[n] is the PROJECTION of the n_th input TSet vector upon the eigenvector
         defined by the the synaptic weight vector W[], after the respective
         neuron convergence */
}NEURON;


typedef struct Network {

char *ParameterFile; /* Input parameters file */
char *InputTSetFileName; /* Training Data Vector Set input file */
unsigned NNr; /* Number of Neurons = Number of Subspaces into which the training set will be decomposed */
unsigned NSn; /* Number of Synapes per Neuron = Number of Input Nodes = Subspaces dimension */
unsigned NSp;  /* Number of vectors (sample vectors) in training set, each one of dimension NSn */
unsigned NumNeurSimTr; /* Number of neurons under simultaneous training */
float GNt;  /* Neurons global (initial) Learning Rate */
float *Nt;  /* neurons learning rate array */
float Alpha; /* Ratio Alpha = [(Eigenvalue)^-1]/LearningRate (Note: A lower Alpha value yields a higher convergence speed - usually  500 < Alpha < 2000 ) */
/* Note: If Alpha = 0 then the neurons learning rate are kept constant (they are not adaptively updated according to the instant eigenvalue), however they can
         be manually changed along training by invoking ChangeParameterMenu() */
NEURON *N; /* neuron array (size=NNr) */
float *Lbd; /* EigenValues array (size=NNr) */
float *Avg;  /* Training Set average vector (size=NSn) */
unsigned LastCvNeur; /* Last converged neuron */
char *buffer; /* file I/O speed up buffer address */
float **TSet; /* Training Set address */
float ConvPercNrmChgCriterion; /* Convergence criterion C ( C is the percent synapses vector squared Euclidean norm change value, below which the neuron synapses vector is regarded as converged to an eigenvector) */
unsigned MaxNumEpoch; /* Max number of training epochs */
float SynIR; /* The value R which defines the synapses random init interval [-R,R] (Note: R is further averaged by the number NSn of synapses per neuron) */

}NEURALNET;


/***************************************************************************
*
* MACROS:
*
****************************************************************************/


static double sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)

static double aulimarg; /* absolute maximum upper limit */
#define AULIM(a) ( fabs(aulimarg=(a)) >= MAXFLOAT ? ( aulimarg< 0.0 ? (-MAXFLOAT) : (MAXFLOAT)) : aulimarg )



/***************************************************************************
*
* FUNCTION PROTOTYPES:
*
****************************************************************************/
void InitNet(NEURALNET *Net);
float RandGen(float Range);
void Quit(EVENT Evt, char *Message);
void ShowNet(NEURALNET *Net);
unsigned HNNTrainer(NEURALNET *Net);
void TSetDeAvg (NEURALNET *Net);
void EigenValSet(NEURALNET *Net);
void InputVectorsProjections(NEURALNET *Net);
void RebuildInputDataSet(NEURALNET *Net,float  **RebuiltTSet);
void SetTitle(void);
void OptionsMenu(void);
int ChangeParameterMenu(NEURALNET *Net,char ch);
void NetWrite(NEURALNET *Net,char *File);
void NetRead(NEURALNET *Net,char *File);
inline void RowShuffle(float  **Mat, unsigned NRows);
void ShowEigenvals(NEURALNET *Net);
double EigenValK(NEURALNET *Net, unsigned k);
void Randomize(void);
float **FloatMatCAlloc(unsigned NRows,unsigned NCols);
void FloatMatFree(float **M);
unsigned long GetFileNumLines(FILE *Fp);
long GetFileNumSamples(FILE *Fp);
float  **FloatMatReadCheck(float **M,char *File,char *buffer,unsigned *NRows,unsigned *NCols, char AllocFlag);
void FloatMatWrite(float **M,char *File,char *buffer,unsigned NRows,unsigned NCols);
void InitParamFileReader(PARFR *Pfr,char *ParameterFile,unsigned *FloatLines,unsigned *AlphaLines, unsigned NumFloatLines, unsigned NumAlphaLines);
void ParamStorageID(PARFR *Pfr, unsigned LineNumber, unsigned *StorageIndex, char *Type);
unsigned long GetSheetNumLines(FILE *Fp, unsigned *MaxLength);
char **CharMatCAlloc(unsigned long NRows,unsigned NCols);
unsigned NumberOfFloatsInLine(char *Line);
float *GetFloatVectFromLine(char *Line, unsigned *VectSize, float *Vector);
void Automatic_ParamFileReader(PARFR *Pfr,char *ParameterFile, char AllocFlag);
void Help(void);

/***************************************************************************
*
* GLOBAL VARIABLES:
*
****************************************************************************/

#ifdef DEBUG_FILE
FILE *Dbg; /* debug file */
#endif

/***************************************************************************
*
*				MAIN():
*
****************************************************************************/
void main (int argc, char *argv[])
{

NEURALNET Net;
char *buffer;

PARFR Pfr;
unsigned Index;
char Type;

char *RebuiltDataSetOutFileName;
char *ConvergedNNSpecOutFileName;

unsigned NumEpoch;

/* parameter file lines specification */
unsigned AlphaLines[NALPHALINES]={0,1,2}; /* see hpar.txt */
unsigned FloatLines[NFLOATLINES]={3,4,5,6,7,8,9};/* see hpar.txt */


/* Tests command line: */
if(argc!=2)Help();

#ifdef DEBUG_FILE
if((Dbg=fopen(DEBUG_FILE,"wt"))==NULL)Quit(WRITE_OPEN_ERR,`DEBUG_FILE);
#endif

/* Init random generator */
Randomize();

/* Set Title */
SetTitle();
/* Set screen */
OptionsMenu();


/* Alloc memory to disk I/O speed-up buffer */
buffer=(char *)malloc(BUFSIZE*sizeof(char));
if(!buffer)Quit(MEMORY_ERR,"");


/* Read Parameters File */

/* initialize  ParamFileReader */
InitParamFileReader(&Pfr,argv[1],FloatLines,AlphaLines,NFLOATLINES,NALPHALINES);

/* identify the parameter/vector at the i_th line of the parameter file,
 i=0,1,...,NALPHALINES+NFLOATLINES-1 and assign it to the corresponding
 variable */
ParamStorageID(&Pfr,0,&Index,&Type);
Net.InputTSetFileName=Pfr.AlphaDat[Index]; /* 0 Training Data Vector Set (Input File) */
ParamStorageID(&Pfr,1,&Index,&Type);
RebuiltDataSetOutFileName=Pfr.AlphaDat[Index]; /* 1 Rebuilt Data Set from the S sub-spaces (Output File) */
ParamStorageID(&Pfr,2,&Index,&Type);
ConvergedNNSpecOutFileName=Pfr.AlphaDat[Index]; /* 2 Converged Neural Net specification (Output File) */
ParamStorageID(&Pfr,3,&Index,&Type);
Net.NNr=Pfr.FloatDat[Index][0]; /* 3 The desired number S of sub-spaces used to decompose the input Data Set (= number of neurons in the Neural Net) */
ParamStorageID(&Pfr,4,&Index,&Type);
Net.GNt=Pfr.FloatDat[Index][0];  /* 4 Initial Global Learning Rate */
ParamStorageID(&Pfr,5,&Index,&Type);
Net.NumNeurSimTr=Pfr.FloatDat[Index][0];  /* 5 Training window size (= number of neurons simultaneously trained) */
ParamStorageID(&Pfr,6,&Index,&Type);
Net.Alpha=Pfr.FloatDat[Index][0]; /* 6 Ratio Alpha = [(Eigenvalue)^-1]/LearningRate (Note: A lower Alpha value yields a higher convergence speed - usually  500 < Alpha < 2000 ) */
ParamStorageID(&Pfr,7,&Index,&Type);
Net.ConvPercNrmChgCriterion=Pfr.FloatDat[Index][0]; /* 7 Convergence criterion C ( C is the percent synapses vector squared Euclidean norm change value, below which the neuron synapses vector is regarded as converged to an eigenvector). */
ParamStorageID(&Pfr,8,&Index,&Type);
Net.MaxNumEpoch=Pfr.FloatDat[Index][0]; /* 8 Max number of training epochs */
ParamStorageID(&Pfr,9,&Index,&Type);
Net.SynIR=Pfr.FloatDat[Index][0]; /* 9 The value R which defines the synapses random init interval [-R,R] (Note: R is further averaged by the number of synapses per neuron) */

/* Define Net Parameters */
Net.ParameterFile=argv[1];
Net.buffer=buffer;

/* Read Training Set from disk */
Net.TSet=FloatMatReadCheck(Net.TSet=NULL,Net.InputTSetFileName,Net.buffer,&Net.NSp,&Net.NSn,1);

/* check and correct */
if(Net.NumNeurSimTr<1)Net.NumNeurSimTr=1;

/* Initialize Net */
InitNet(&Net);

/* Subtracts statistical average from Training Set. Stores average in Net. */
TSetDeAvg(&Net);

/* Hebbian Neural Network Trainer */
NumEpoch=HNNTrainer(&Net);

/* Determine the eigenvalue set */
EigenValSet(&Net);

/* Show Network */
textcolor(LIGHTGRAY);system("cls");
ShowNet(&Net);
printf("\nTotal number of training epochs = %u\n",NumEpoch);


/* Determine the projections of each TSet vector upon each axis of the orthonormal basis.
Each axis is defined by one eigenvector. Each eigenvector is a synaptic weights vector
after the convergence of its respective neuron. */
InputVectorsProjections(&Net);

/* Rebuild the entire set of input vectors from its NNr sub-spaces */
RebuildInputDataSet(&Net,Net.TSet);

/* Write the rebuilt TSet to file RebuiltDataSetOutFileName */
FloatMatWrite(Net.TSet,RebuiltDataSetOutFileName,Net.buffer,Net.NSp,Net.NSn);

/* Writes Net to file ConvergedNNSpecOutFileName */
NetWrite(&Net,ConvergedNNSpecOutFileName);

/* beeps */
putch(BEL);

}



/***************************************************************************
* FUNC: void InitNet(NEURALNET *Net)
*
* DESC: Init NEURALNET Net.
****************************************************************************/
void InitNet(NEURALNET *Net)
{

register unsigned n; /* n indexes Neuron */

Net->N=(NEURON *)calloc((unsigned)Net->NNr,sizeof(NEURON));
if(!Net->N)Quit(MEMORY_ERR,"");

Net->Nt=(float *)calloc((unsigned)Net->NNr,sizeof(float));
if(!Net->Nt)Quit(MEMORY_ERR,"");

Net->Lbd=(float *)calloc((unsigned)Net->NNr,sizeof(float));
if(!Net->Lbd)Quit(MEMORY_ERR,"");

Net->Avg=(float *)calloc((unsigned)Net->NSn,sizeof(float));
if(!Net->Avg)Quit(MEMORY_ERR,"");



/* for each neuron n, alloc memory to synapses weights vector and to converged output array */
for(n=0;n<Net->NNr;n++){

Net->N[n].W=(double *)calloc((unsigned)Net->NSn,sizeof(double));
if(!Net->N[n].W)Quit(MEMORY_ERR,"");

Net->N[n].yc=(float *)calloc((unsigned)Net->NSp,sizeof(float));
if(!Net->N[n].yc)Quit(MEMORY_ERR,"");


/* Set global learning rate to each individual neuron learning rate */
Net->Nt[n]=Net->GNt;

}/* for n */


}



/***************************************************************************
* FUNC: float RandGen(float Range)
*
* DESC: RandGen() generates random numbers in the range {-Range, +Range} with
*       uniform probabilistic distribution.
*
****************************************************************************/
float RandGen(float Range)
{

return  (((float)rand()/(RAND_MAX/2))-1)*Range;

}

/***************************************************************************
* FUNC: void ShowNet(NEURALNET *Net)
*
* DESC: ShowNet() show NEURALNET Net in screen.
****************************************************************************/
void ShowNet(NEURALNET *Net)

{

register unsigned n,s; /* n indexes Neuron, s indexes Synapses */

float Nrm;


#ifdef SHOWEIGENVEC
printf("\n\nSynapses Values (EigenVectors):");

for(n=0;n<Net->NNr;n++){
printf("\nV%u: ",n);
for(s=0;s<Net->NSn;s++)printf("%5.2g\t",Net->N[n].W[s]);
}/* for n */
#endif


printf("\n\n\tEigenvalues:\t\tEigenvectors squared Euclidean Norm:\n");


for(n=0;n<Net->NNr;n++){

	Nrm=0;

	/* determine the squared Euclidean norm of synapses weights vector for neuron n */
	for(s=0;s<Net->NSn;s++){
	Nrm+=SQR(Net->N[n].W[s]);
	}/* for s */

printf("L%u:\t%8.6g\t\t\t%6.3g\n",n,Net->Lbd[n],Nrm);
}/* for n */


}


/*****************************************************************************
* FUNC: void Quit(EVENT Evt, char *Message)
*
* DESC: Print Event message and exit to shell.
******************************************************************************/
void Quit(EVENT Evt, char *Message)
{

static char *Event[] = {

/*  NO_EVENT */       " ",
/*  MEMORY_ERR */     "Not enough memory",
/*  READ_OPEN_ERR */  "I can't find the file",
/*  READ_ERR */       "Error reading file",
/*  WRITE_OPEN_ERR */ "I can't open the file",
/*  WRITE_ERR */      "I can't write the file",
/*  BUFFER_ERR */     "I can't attach buffer to file",
/*  SIZE_ERR */       "Specified size doesn't match data size in file",
/*  MAXIT_ERR */      "Exceeded max number of iterations in function",
/*  LOGIC_ERR */      "Logic error in function",
/*  READ_NNET_ERR */  "Incompatible neural network in file",
/*  MTXDIM_ERR */     "Invalid matrix dimensions in file",
/*  PARF_ERR  */      "Wrong parameter file",
/*  GFTD_ERR  */      "I can't get time & date info from file",
/*  MAXFLOAT_READ_ERR */"Exceeded float maximum absolute value reading file",
/*---------*/
/* NET_SIZE_ERR */    "Read incompatible neural network from file"
};

	system("cls");

	if (Evt != NO_EVENT) {


		if(Message[0]){
		fprintf (stderr,"%s %s!\n", Event[Evt], Message);
                #if(0)
		putch(BEL);
                #endif
		exit (Evt);
		}
		else{
		fprintf (stderr,"%s!\n", Event[Evt]);
                #if(0)
		putch(BEL);
                #endif
		exit (Evt);
		}

	}
	exit (0);
}




/***************************************************************************
* FUNC: unsigned HNNTrainer(NEURALNET *Net)
*
* DESC: Hebbian Neural Network Trainer.
****************************************************************************/
unsigned HNNTrainer(NEURALNET *Net)
{

#define NUMTAB 5 /* number of screen columns showing the eigenvectors squared Euclidean norm */

unsigned NumEpoch,Epoch=0;
double Nrm,PrevNrm=0,PercentNrmChg=0,AvgPercNrmChg;
double PercNrmChgHistory[NAVGITV];
double PresentEgv;

double deflat;

register unsigned n,s,k;
register unsigned long It;
unsigned AvNrmChgChkItv;

unsigned long MaxNumIter=(long)(Net->NSp)*Net->MaxNumEpoch;
unsigned IterChkItrvl=(Net->NSp)*PNCC;
 

char FNotCvrg;

unsigned long mdl;

unsigned Tab[NUMTAB]={1,17,33,49,65}; /* neurons squared Euclidean norm screen NUMTAB column x_coordinates */
unsigned ShwCvrgCount=0;
unsigned TabIdx;
char ch;
unsigned sx,sy,yindex;

/* Init percent squared Euclidean norm change history array */
for(n=0;n<NAVGITV;n++)PercNrmChgHistory[n]=10*Net->ConvPercNrmChgCriterion;

/* Randomly init neurons synapse vector and zeroes neurons outputs */
for(n=0;n<Net->NNr;n++){
Net->N[n].y=0;

for(s=0;s<Net->NSn;s++){
Net->N[n].W[s]=(RandGen(Net->SynIR))/(Net->NSn);
}/* for s */


for(k=0;k<Net->NSp;k++){
Net->N[n].yc[k]=0;
}/* for k */

}/* for n */


/* init Last converged neuron */
Net->LastCvNeur=0;

/* init averaged squared Euclidean norm change check interval */
AvNrmChgChkItv=0;

/* iteration loop */
for(It=0;It<MaxNumIter;It++){

/* modulus Net->NSp */
mdl=It%(Net->NSp);

/* determine the output of each neuron */
for(n=0;n<Net->LastCvNeur+Net->NumNeurSimTr;n++){
if(n>=Net->NNr)break;

Net->N[n].y=0;
	for(s=0;s<Net->NSn;s++){
	Net->N[n].y+=Net->N[n].W[s]*Net->TSet[(unsigned)mdl][s];
	}/* for s */
Net->N[n].y=AULIM(Net->N[n].y); /* to avoid eventual float overflow */
}/* for n */


/* update synapses weights */
for(n=Net->LastCvNeur;n<Net->LastCvNeur+Net->NumNeurSimTr;n++){
if(n>=Net->NNr)break;

	for(s=0;s<Net->NSn;s++){


		deflat=0;
		for(k=0;k<=n;k++){
		deflat+=Net->N[k].W[s]*Net->N[k].y;
        	}/* for k */
                deflat=AULIM(deflat); /* to avoid eventual float overflow */

Net->N[n].W[s]+=Net->Nt[n]*Net->N[n].y*(Net->TSet[(unsigned)mdl][s]-deflat);
Net->N[n].W[s]=AULIM(Net->N[n].W[s]);/* to avoid eventual float overflow */
	}/* for s */

}/* for n */



/* Checks for each neuron convergence at each IterChkItrvl */
if(!(It%(IterChkItrvl))&&It>0){

FNotCvrg=1;


/* Call changing parameters menu if keyboard hit */
if(kbhit()){
ch=toupper(getch());
if(ChangeParameterMenu(Net,ch)<0)FNotCvrg=0;
SetTitle();OptionsMenu();
}/* if(kbhit()) */



/* Show convergence in screen */
//window(1,5,80,24);
sx=1; sy=5;
for(yindex=sy; yindex<=24; yindex++){
gotoxy(sx,yindex);clreol();
}

for(ShwCvrgCount=0;ShwCvrgCount<Net->NNr;ShwCvrgCount++){
gotoxy(sx,sy);
if(ShwCvrgCount>Net->LastCvNeur+Net->NumNeurSimTr-1)textcolor(LIGHTGREEN);
else if(ShwCvrgCount>Net->LastCvNeur)textcolor(YELLOW);
else textcolor(LIGHTBLUE);
TabIdx=ShwCvrgCount%NUMTAB;
gotoxy(Tab[TabIdx],wherey());


Nrm=0;

	/* determine the squared Euclidean norm of neuron n */
	for(s=0;s<Net->NSn;s++){
	Nrm+=SQR(Net->N[ShwCvrgCount].W[s]);
	}/* for s */


if(ShwCvrgCount==Net->LastCvNeur){
	textcolor(LIGHTMAGENTA);
	Epoch=(unsigned)(It/(Net->NSp));
	if(Epoch==1)PrevNrm=Nrm;
	PercentNrmChg=100*((Nrm/PrevNrm)-1);
	PrevNrm=Nrm;

}


cprintf("N%u:%.3f",ShwCvrgCount,Nrm);
if(TabIdx==NUMTAB-1)cprintf("\n\r");
sx=wherex();sy=wherey();
}/* for ShwCvrgCount */



/* update the percent squared Euclidean norm change history array and determine the percent squared Euclidean norm change average */
PercNrmChgHistory[AvNrmChgChkItv++]=fabs(PercentNrmChg);
if(AvNrmChgChkItv>=NAVGITV)AvNrmChgChkItv=0;
AvgPercNrmChg=0.0;
for(s=0;s<NAVGITV;s++)AvgPercNrmChg+=PercNrmChgHistory[s];
AvgPercNrmChg/=NAVGITV;


/* Adaptively update the Learning Rate if Net->Alpha>0  */
PresentEgv=EigenValK(Net,Net->LastCvNeur);
if((Net->LastCvNeur>0)&&(Net->Alpha>0)){
Net->GNt=(1/(PresentEgv*Net->Alpha));
for(s=0;s<Net->NNr;s++){Net->Nt[s]=Net->GNt;}
}

/* Show bottom status bar */
//window(1,25,80,25);
gotoxy(1,25);clreol();
textcolor(LIGHTBLUE);
cprintf("Epoch:%u   ",Epoch);
textcolor(LIGHTMAGENTA);
//cprintf("Training Range:%u-%u   ",Net->LastCvNeur,Net->LastCvNeur+Net->NumNeurSimTr-1);
cprintf("Eigenvalue=%5.3g   ",PresentEgv);
cprintf("Sq.Eucl.NormChg: %+5.3g %c   ",PercentNrmChg,'%');
cprintf("Lrn.Rate: %5.3g",Net->Nt[Net->LastCvNeur]);



/* Check convergence due to positive small enough averaged percent rate of squared Euclidean norm change */
if(AvgPercNrmChg<Net->ConvPercNrmChgCriterion){
	if(PercentNrmChg>0){
	Net->LastCvNeur++;
	for(s=0;s<NAVGITV;s++)PercNrmChgHistory[s]=10*Net->ConvPercNrmChgCriterion;
	}
}

if(Net->LastCvNeur>=Net->NNr){
FNotCvrg=0;
Net->LastCvNeur--;
}


/* if found,valid eigenvectors, return */
if(!FNotCvrg){
NumEpoch=(unsigned)(It/(Net->NSp));
return NumEpoch;
}


}/* if(!(It%IterChkItrvl)&&It>0) */

/* If it is the end of an epoch, shuffle TSet */
if(!(It%(Net->NSp)))RowShuffle(Net->TSet,Net->NSp);

}/* for It */


NumEpoch=(unsigned)(It/(Net->NSp));
return NumEpoch;


}


/***************************************************************************
* FUNC: void TSetDeAvg (NEURALNET *Net)
*
* DESC: Subtracts the average vector Avg[] from training set TSet.
*       Stores Avg[] in Net.
****************************************************************************/
void TSetDeAvg (NEURALNET *Net)
{

register unsigned i,j;



for(j=0;j<Net->NSn;j++){
Net->Avg[j]=0;

for(i=0;i<Net->NSp;i++){
Net->Avg[j]+=Net->TSet[i][j];
}

Net->Avg[j]/=Net->NSp;

for(i=0;i<Net->NSp;i++){
Net->TSet[i][j]-=Net->Avg[j];
}


}



}


/***************************************************************************
* FUNC: void EigenValSet(NEURALNET *Net)
*
* DESC: Determine the eigenvalue set and store it in Net.
****************************************************************************/
void EigenValSet(NEURALNET *Net)
{

register unsigned i,j,k;

float lambda;
double acc;


for(k=0;k<Net->NNr;k++){

lambda=0;

for(i=0;i<Net->NSp;i++){

acc=0;

for(j=0;j<Net->NSn;j++){
acc+=Net->TSet[i][j]*Net->N[k].W[j];
}/* for j */

lambda+=SQR(acc);

}/* for i */


Net->Lbd[k]=lambda/Net->NSp;

} /* for k */


}


/***************************************************************************
* FUNC: double EigenValK(NEURALNET *Net, unsigned k)
*
* DESC: Determine the k_th eigenvalue.
****************************************************************************/
double EigenValK(NEURALNET *Net, unsigned k)
{

register unsigned i,j;

double lambda;
double acc;



lambda=0;

for(i=0;i<Net->NSp;i++){

acc=0;

for(j=0;j<Net->NSn;j++){
acc+=Net->TSet[i][j]*Net->N[k].W[j];
}/* for j */

lambda+=SQR(acc);

}/* for i */


return lambda/Net->NSp;


}




/***************************************************************************
* FUNC: void RebuildInputDataSet(NEURALNET *Net,float  **RebuiltTSet)
*
* DESC: Rebuild the entire set of vectors from its NNr sub-spaces.
*
*
* NOTE1: yc[n] is the PROJECTION of the n_th input TSet vector upon the eigenvector
*        defined by the the synaptic weight vector W[], after the respective
*        neuron convergence.
*
* NOTE2: Each eigenvector define an axis of an orthonormal basis which aligns
*        with the highest variance directions of the input vector set TSet.
****************************************************************************/
void RebuildInputDataSet(NEURALNET *Net,float  **RebuiltTSet)

{
unsigned n,k,i,j;
double Tmp;


/* adds average */
for(i=0;i<Net->NSp;i++){
for(j=0;j<Net->NSn;j++){
RebuiltTSet[i][j]=Net->Avg[j];
}
}


/* rebuild input vector set from its projections (sub-spaces) */
for(n=0;n<Net->NSp;n++){


for(k=0;k<((Net->NNr));k++){


	for(j=0;j<Net->NSn;j++){
        Tmp=AULIM(Net->N[k].yc[n]*Net->N[k].W[j]);
        if(fabs(Tmp>=MAXFLOAT))RebuiltTSet[n][j]=Tmp;
	else RebuiltTSet[n][j]+=Tmp;
	}/* for J */


}/* for k */



}/* for n */


}


/***************************************************************************
* FUNC: void InputVectorsProjections(NEURALNET *Net)
*
* DESC: Determine the projections of each TSet vector upon the orthonormal
*       basis defined by the set of eigenvectors (synaptic weights vector,
*       after convergence).
*
* NOTE: Each eigenvector define an axis of an orthonormal basis which aligns
*       with the highest variance directions of the input vector set TSet.
*       The sum of the TSet projections upon one axis define one sub-space.
****************************************************************************/
void InputVectorsProjections(NEURALNET *Net)
{

register unsigned i,n,s;

/* Re-read Training Set from disk - to circumvent RowShuffle() TSet vectors random ordering */
Net->TSet=FloatMatReadCheck(Net->TSet,Net->InputTSetFileName,Net->buffer,&Net->NSp,&Net->NSn,0);

/* de-averages TSet */
TSetDeAvg (Net);

for(i=0;i<Net->NSp;i++){

/* determine the output of each neuron */
for(n=0;n<Net->NNr;n++){
Net->N[n].y=0;
	for(s=0;s<Net->NSn;s++){
	Net->N[n].y+=AULIM(Net->N[n].W[s]*Net->TSet[i][s]);
	}/* for s */

Net->N[n].yc[i]=Net->N[n].y;

}/* for n */


}/* for i */


}


/**************************************************************************
* FUNC: void OptionsMenu(void)
*
* DESC: Draw options menu in screen.
*************************************************************************/
void OptionsMenu(void)
{

//window(1,3,80,4);clrscr();

gotoxy(1,4);clreol();
gotoxy(1,3);clreol();

textcolor(WHITE);cprintf("L");textcolor(LIGHTRED);cprintf("earning Rate  ");
textcolor(WHITE);cprintf("R");textcolor(LIGHTRED);cprintf("ead Net  ");
textcolor(WHITE);cprintf("W");textcolor(LIGHTRED);cprintf("rite Net  ");
textcolor(WHITE);cprintf("E");textcolor(LIGHTRED);cprintf("igenvalues  ");
textcolor(WHITE);cprintf("T");textcolor(LIGHTRED);cprintf("raining Range  ");
textcolor(LIGHTRED);cprintf("Re");textcolor(WHITE);cprintf("s");textcolor(LIGHTRED);cprintf("tart  ");
textcolor(WHITE);cprintf("Q");textcolor(LIGHTRED);cprintf("uit");

}


/**************************************************************************
* FUNC: void SetTitle(void)
*
* DESC: Draw title
*************************************************************************/
void SetTitle(void)
{

//window(1,1,80,2);
gotoxy(1,2);clreol();
gotoxy(1,1);clreol();
textcolor(LIGHTCYAN);
gotoxy(24,wherey());cprintf("Hebbian Neural Network Trainer");

}




/**************************************************************************
* FUNC: int ChangeParameterMenu(NEURALNET *Net,char ch)
*
* DESC: Change Neural Net parameters "on fly". Return -1 on quit procedure,
*       or else return 0.
*************************************************************************/
int ChangeParameterMenu(NEURALNET *Net,char ch)
{

#define CGETS_NCH 20 /* max number of chars read by cgets() */

char str[CGETS_NCH+3];
char *p;
unsigned n;
register unsigned s;


str[0]=CGETS_NCH+1;  /* There's space for CGETS_NCH chars plus the NULL terminator */

//window(1,3,80,4);clrscr();

gotoxy(1,4);clreol();
gotoxy(1,3);clreol();

switch(ch){

case 'L':
clrscr();
textcolor(WHITE);cprintf("G");textcolor(LIGHTRED);cprintf("lobal Lrn.Rate ");
textcolor(WHITE);cprintf("N");textcolor(LIGHTRED);cprintf("euron Lrn.Rat  ");
textcolor(WHITE);cprintf("R");textcolor(LIGHTRED);cprintf("atio [(Eigenval)^-1]/[Lrn.Rate]");
ch=toupper(getch());

	switch(ch){

	case 'G':
	clrscr();
	textcolor(YELLOW);cprintf("Global Learning Rate = %-8.6g  ",Net->GNt);
	textcolor(LIGHTRED);cprintf("Enter new value:");
	p = cgets(str);
	sscanf(p,"%f",&Net->GNt);
	for(n=0;n<Net->NNr;n++){Net->Nt[n]=Net->GNt;}
	break;

	case 'N':
	LBL00:
	clrscr();
	textcolor(YELLOW);cprintf("Change Learning Rate of which neuron? ");
	p = cgets(str);
	sscanf(p,"%u",&n);
	if(n>=Net->NNr){putch(BEL);goto LBL00;}
	clrscr();
	textcolor(YELLOW);cprintf("Neuron %u Learning Rate = %-8.6g  ",n,Net->Nt[n]);
	textcolor(LIGHTRED);cprintf("Enter new value:");
	p = cgets(str);
	sscanf(p,"%f",&Net->Nt[n]);
	break;

	case 'R':
	clrscr();
	textcolor(YELLOW);cprintf("Ratio [(Eigenval)^-1]/[Lrn.Rate] = %-8.6g  ",Net->Alpha);
	textcolor(LIGHTRED);cprintf("Enter new value:");
	p = cgets(str);
	sscanf(p,"%f",&Net->Alpha);
	break;
	}

break;

case 'R':
textcolor(YELLOW);cprintf("Read Network from what file? ");
textcolor(LIGHTRED);
p = cgets(str);
if(!*p)break;
NetRead(Net,p);
//NetWrite(Net,"test.txt");
break;

case 'W':
textcolor(YELLOW);cprintf("Write Network to what file? ");
textcolor(LIGHTRED);
p = cgets(str);
if(!*p)break;
EigenValSet(Net); /* Determine the eigenvalue set */
/* Determine the projections of each TSet vector upon each axis of the
orthonormal basis defined by the eigenvector set: */
InputVectorsProjections(Net);
NetWrite(Net,p);
break;

case 'E':
ShowEigenvals(Net);
break;

case 'S':

LBL01:clrscr();
textcolor(YELLOW);cprintf("Are you sure you want to restart training?");cscanf("%c",&ch);
if(toupper(ch)=='Y'){
clrscr();
textcolor(YELLOW);cprintf("Restart training from which neuron? ");
p = cgets(str);
sscanf(p,"%u",&n);
if(n>=Net->NNr){putch(BEL);goto LBL01;}
clrscr();
textcolor(YELLOW);cprintf("Initialize synapses weights ramdomly? ");
cscanf("%c",&ch);
if(toupper(ch)=='Y'){
for(s=0;s<Net->NSn;s++){
Net->N[n].W[s]=(RandGen(Net->SynIR))/(Net->NSn);
}/* for s */
}

Net->LastCvNeur=n;
}

break;


case 'T':
clrscr();
textcolor(YELLOW);cprintf("Number of neurons simultaneously trained = %u  ",Net->NumNeurSimTr);
textcolor(LIGHTRED);cprintf("Enter new value:");
p = cgets(str);
sscanf(p,"%u",&Net->NumNeurSimTr);
if(Net->NumNeurSimTr<1)Net->NumNeurSimTr=1;
break;


case 'Q':
textcolor(YELLOW);cprintf("Are you sure you want to quit?");cscanf("%c",&ch);
if(toupper(ch)=='Y')return -1;;
break;

}

return 0;

}



/**************************************************************************
* FUNC: void NetWrite(NEURALNET *Net,char *File)
*
* DESC: Write the neural network specification to the ascii file File.
*************************************************************************/
void NetWrite(NEURALNET *Net,char *File)
{

FILE *fp;

char ParFileTimeDate[128];

register unsigned i,j;


/* get date & time info of Net->ParameterFile */
//if(GetFileTimeData(Net->ParameterFile,ParFileTimeDate))Quit(GFTD_ERR,Net->ParameterFile);


if((fp=fopen(File,"wt"))==NULL)Quit(WRITE_OPEN_ERR,File);
if (setvbuf(fp,Net->buffer,_IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,File);


fprintf(fp,"%8s\t%c HNNTR parameter file and its date & time.\n","ParFileTimeDate",'%');
fprintf(fp,"%8u\t%c The number of vectors in the input training set.\n",Net->NSp,'%');
fprintf(fp,"%8u\t%c The number of desired sub-spaces (=number of neurons=number of eigenvalues & eigenvectors).\n",Net->NNr,'%');
fprintf(fp,"%8u\t%c The dimension of each sub-space (=input vectors dimension=number of synapses per neuron).\n",Net->NSn,'%');
fprintf(fp,"%8u\t%c The last neuron that has converged to an eigenvector.\n",Net->LastCvNeur,'%');
fprintf(fp,"%8.6g\t%c The final learning rate.\n",Net->GNt,'%');


fprintf(fp,"The input training set average vector:\n");
for(i=0;i<Net->NSn;i++)fprintf(fp,"%8.6g\t",Net->Avg[i]);
fprintf(fp,"\n");


fprintf(fp, "The %u eigenvalues (the variance of each one of the %u sub-spaces):\n",Net->NNr,Net->NNr);
for(i=0;i<Net->NNr;i++)fprintf(fp,"%8.6g\t",Net->Lbd[i]);
fprintf(fp,"\n");


fprintf(fp,"The %u corresponding %u-dimensional eigenvectors (orthonormal basis axes unitary vectors):\n", Net->NNr,Net->NSn);


for(i=0;i<Net->NNr;i++){
for(j=0;j<Net->NSn;j++){

fprintf(fp,"%8.6g\t",Net->N[i].W[j]);

}
fprintf(fp,"\n");
}


fprintf(fp,"The output of each one of the %u neurons at the instant the Neural Net was saved:\n",Net->NNr);
for(i=0;i<Net->NNr;i++)fprintf(fp,"%8.6g\t",Net->N[i].y);
fprintf(fp,"\n");


fprintf(fp,"The projection of each input vector upon each one of the %u axes of the orthonormal basis:\n",Net->NNr);


for(i=0;i<Net->NSp;i++){
for(j=0;j<Net->NNr;j++){

fprintf(fp,"%8.6g\t",Net->N[j].yc[i]);

}
fprintf(fp,"\n");
}

fclose(fp);

}


/**************************************************************************
* FUNC: void NetRead(NEURALNET *Net,char *File)
*
* DESC: Read the neural network specification from the ascii file File.
*************************************************************************/
void NetRead(NEURALNET *Net,char *File)
{

register unsigned i,j,LineCount;

static PARFR PfrS;
static char EntryFlag;
char AllocFlag;


unsigned NNr; /* Number of Neurons = Number of Subspaces into which the training set will be decomposed */
unsigned NSn; /* Number of Synapes per Neuron = Number of Input Nodes = Subspaces dimension */
unsigned NSp;  /* Number of samples vectors in training set, each one of dimension NSn */


/*  AllocFlag=1 on the first call of NetRead(). AllocFlag=0 in
any other instant */
if(!EntryFlag){EntryFlag=AllocFlag=1;}
else {AllocFlag=0;}


/* read neural network from File */
Automatic_ParamFileReader(&PfrS,File,AllocFlag);


/* Get architecture parameters */
LineCount=0;

NSp=PfrS.FloatDat[LineCount++][0]; /* LineCount = 0 */
NNr=PfrS.FloatDat[LineCount++][0]; /* LineCount = 1 */
NSn=PfrS.FloatDat[LineCount++][0]; /* LineCount = 2 */

/* check architecture compatibility */
if(Net->NSp!=NSp||Net->NNr!=NNr||Net->NSn!=NSn)Quit(NET_SIZE_ERR,File);

/* Get the rest of the parameters */
Net->LastCvNeur=PfrS.FloatDat[LineCount++][0]; /* LineCount = 3 */
Net->GNt=PfrS.FloatDat[LineCount++][0]; /* LineCount = 4 */

for(i=0;i<Net->NSn;i++){
Net->Avg[i]=PfrS.FloatDat[LineCount][i]; /* LineCount = 5 */
}
LineCount++;

for(i=0;i<Net->NSn;i++){
Net->Lbd[i]=PfrS.FloatDat[LineCount][i]; /* LineCount = 6 */
}
LineCount++;

for(i=0;i<Net->NNr;i++){
for(j=0;j<Net->NSn;j++){
Net->N[i].W[j]=PfrS.FloatDat[LineCount][j]; /* LineCount = 7 to 7+(NNr-1) */
}
LineCount++;
}

for(i=0;i<Net->NNr;i++){
Net->N[i].y=PfrS.FloatDat[LineCount][i]; /* LineCount = 7+NNr */
}
LineCount++;


for(i=0;i<Net->NSp;i++){
for(j=0;j<Net->NNr;j++){
Net->N[j].yc[i]=PfrS.FloatDat[LineCount][j]; /* LineCount = 8+NNr to 8+NNr + (NSp-1) */
}
LineCount++;
}



}



/***************************************************************************
* FUNC: inline void RowShuffle(float  **Mat, unsigned NRows)
*
* DESC: Randomly shuffles the rows of matrix Mat.
****************************************************************************/
inline void RowShuffle(float  **Mat, unsigned NRows)
{

register unsigned i;
unsigned r1;
unsigned r2;
float  *trow;

for(i=0;i<NRows;i++)
{

r1=rand()%NRows;
r2=rand()%NRows;


trow=Mat[r1];
Mat[r1]=Mat[r2];
Mat[r2]=trow;

}

}

/***************************************************************************
* FUNC: void ShowEigenvals(NEURALNET *Net)
*
* DESC: Show eigenvals in screen.
****************************************************************************/
void ShowEigenvals(NEURALNET *Net)
{

unsigned NeuronCount;
unsigned Tab[NUMTAB]={1,17,33,49,65}; /* eigenvalues screen NUMTAB column x_coordinates */
unsigned TabIdx;
unsigned sx,sy,yindex;


//window(1,3,80,4);clrscr();
//gotoxy(1,3);clreol();
//gotoxy(1,4);clreol();
//gotoxy(1,3);

clrscr();
textcolor(YELLOW);gotoxy(32,1);cprintf("Eigenvalues:");

//window(1,5,80,24);
sx=1; sy=2;
for(yindex=sy; yindex<=24; yindex++){
gotoxy(sx,yindex);clreol();
}

textcolor(LIGHTMAGENTA);


/* Determine the eigenvalue set */
EigenValSet(Net);

/* Show eigenvalues in screen */
//sx=sy=1;

for(NeuronCount=0;NeuronCount<Net->NNr;NeuronCount++){
gotoxy(sx,sy);

TabIdx=NeuronCount%5;
gotoxy(Tab[TabIdx],wherey());

cprintf("L%u:%8.5g",NeuronCount,Net->Lbd[NeuronCount]);
if(TabIdx==4)cprintf("\n\r");
sx=wherex();sy=wherey();
}/* for NeuronCount */

//window(1,25,80,25);
gotoxy(1,25);clreol();
textcolor(YELLOW);

cprintf("Press any key to continue...");getch();
clrscr();

}

/***************************************************************************
* FUNC: void  Randomize(void)
*
* DESC: Randomize random generator seed.
****************************************************************************/
void Randomize(void)
{
srand((unsigned) time(NULL));
}

/*****************************************************************************
* FUNC: float **FloatMatCAlloc(unsigned NRows,unsigned NCols)
*
* DESC: Alloc memory for a NRows x NCols  float matrix.
*****************************************************************************/
float **FloatMatCAlloc(unsigned NRows,unsigned NCols)
{

float  *x;
float  **Mx;
register unsigned i;


x = (float *)calloc((unsigned long)(NRows*NCols),sizeof(float));
if(!x)Quit(MEMORY_ERR,"");


Mx=(float  **)calloc((unsigned long)NRows,sizeof(float *));
if(!Mx)Quit(MEMORY_ERR,"");


for(i=0;i<NRows;i++){

Mx[i]=(float *)&x[i*NCols];

}

return Mx;
}


/*****************************************************************************
* FUNC: void FloatMatFree(float **M)
*
* DESC: Free memory for the float matrix M.
*****************************************************************************/
void FloatMatFree(float **M)
{

free(M[0]);
free(M);

}


/*****************************************************************************
* FUNC: unsigned long GetFileNumLines(FILE *Fp)
*
* DESC: Return the number of lines in the text file pointed by Fp.
******************************************************************************/
unsigned long GetFileNumLines(FILE *Fp)
{

  char line[MAXLINESIZE];
  register unsigned long Count;


  Count=0;
  while (fgets(line,MAXLINESIZE-2,Fp)){
  Count++;
  }

  if(strlen(line)==1)Count--;/* to discount the last \n in file
                              for the case the user have typed it */


  rewind(Fp);
  return Count;

}



/*****************************************************************************
* FUNC: long GetFileNumSamples(FILE *Fp)
*
* DESC: Return the number of numeric samples in the text file pointed by Fp.
*       Return -1 on read error.
*****************************************************************************/
long GetFileNumSamples(FILE *Fp)
{

register long Count;
float x;
int test;

Count=0;

while((test=fscanf(Fp,"%f",&x))!=-1){
if(test!=1)return -1;
Count++;
}

rewind(Fp);
return Count;

}


/*****************************************************************************
* FUNC: float  **FloatMatReadCheck(float **M,char *file,char *buffer,unsigned *NRows,unsigned *NCols, char AllocFlag)
*
* DESC: Read an ascii float matrix from 'file' and store it in matrix
*       M[NRows][NCols]. If AllocFlag==1 the function allocates memory
*       for M and the arguments *NRows and *NCols identify the size of M.
*
*       FloatMatReadCheck() checks the float compatibility of the read data.
*****************************************************************************/
float  **FloatMatReadCheck(float **M,char *file,char *buffer,unsigned *NRows,unsigned *NCols, char AllocFlag)
{

register unsigned i,j;
FILE *in;
long NumSamples;
unsigned long NumLines;
int Test;
double Value;

if((in=fopen(file,"rt"))==NULL)Quit(READ_OPEN_ERR,file);
if (setvbuf(in, buffer, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,file);

NumSamples=GetFileNumSamples(in);
if(NumSamples==-1)Quit(READ_ERR,file);
NumLines=GetFileNumLines(in);

if(NumSamples%NumLines)Quit(MTXDIM_ERR,file); /* check valid matrix dimensions */

*NRows=NumLines;
*NCols=NumSamples/NumLines;

if(AllocFlag==1)M=FloatMatCAlloc(*NRows,*NCols);

for(i=0;i<*NRows;i++){
for(j=0;j<*NCols;j++){
Test=fscanf(in,"%lf",&Value);
if(Test!=1)Quit(READ_ERR,file);
if(fabs(Value)>MAXFLOAT)Quit(MAXFLOAT_READ_ERR,file);
M[i][j]=Value;
}
}

fclose(in);

return M;
}


/*****************************************************************************
* FUNC: void FloatMatWrite(float **M,char *File,char *buffer,unsigned NRows,unsigned NCols)
*
* DESC: Writes a float NRows x NCols matrix to an ascii file File. If File=NULL, it
*       writes to stdout.
*****************************************************************************/
void FloatMatWrite(float **M,char *File,char *buffer,unsigned NRows,unsigned NCols)
{

register unsigned i,j;
FILE *out;
int test;

if(File!=NULL){
if((out=fopen(File,"wt"))==NULL)Quit(WRITE_OPEN_ERR,File);
if (setvbuf(out, buffer, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,File);
}
else out=stdout;


for(i=0;i<NRows;i++){
for(j=0;j<NCols;j++){

fprintf(out,"%8.6g\t",M[i][j]);

}

test=fprintf(out,"\n");
if(test!=sizeof(char))Quit(WRITE_ERR,File);

}

if(File!=NULL)fclose(out);


}


/*****************************************************************************
* FUNC: void InitParamFileReader(PARFR *Pfr,char *ParameterFile,unsigned *FloatLines,unsigned *AlphaLines, unsigned NumFloatLines, unsigned NumAlphaLines)
*
* DESC: Init parameter file reader.
*****************************************************************************/
void InitParamFileReader(PARFR *Pfr,char *ParameterFile,unsigned *FloatLines,unsigned *AlphaLines, unsigned NumFloatLines, unsigned NumAlphaLines)
{

register unsigned i;

Pfr->ParameterFile=ParameterFile;
Pfr->FloatLines=FloatLines;
Pfr->AlphaLines=AlphaLines;
Pfr->NumFloatLines=NumFloatLines;
Pfr->NumAlphaLines=NumAlphaLines;


/* Alloc memory for disk I/O speed-up buffers */
Pfr->BufferI=(char *)calloc(SMALLBUFSIZE,sizeof(char));
if(!Pfr->BufferI)Quit(MEMORY_ERR,"");


/* Open parameter file */
if((Pfr->IFp=fopen(Pfr->ParameterFile,"rt"))==NULL)Quit(READ_OPEN_ERR,Pfr->ParameterFile);
if (setvbuf(Pfr->IFp, Pfr->BufferI, _IOFBF, SMALLBUFSIZE) != 0)Quit(BUFFER_ERR,Pfr->ParameterFile);

/* Get the total number of lines and the max line length */
Pfr->NumLines=GetSheetNumLines(Pfr->IFp,&Pfr->LineMaxLength);

/* check */
if(Pfr->NumLines!=(Pfr->NumFloatLines+Pfr->NumAlphaLines))Quit(PARF_ERR,Pfr->ParameterFile);

/* Alloc memory for Sheet */
Pfr->Sheet=CharMatCAlloc(Pfr->NumLines,Pfr->LineMaxLength);

/* Read sheet and store it into Sheet */
for(i=0;i<Pfr->NumLines;i++)fgets(Pfr->Sheet[i],Pfr->LineMaxLength,Pfr->IFp);

/* close parameter file */
fclose(Pfr->IFp);

/* Alloc memory for float FloatDat row pointers */
Pfr->FloatDat=(float **)calloc(Pfr->NumFloatLines,sizeof(float *));
if(!Pfr->FloatDat)Quit(MEMORY_ERR,"");

/* Alloc memory for alphanumeric data */
Pfr->AlphaDat=CharMatCAlloc(Pfr->NumAlphaLines,MAXARGLEN);

/* Alloc memory for the array that contains the size of each row of FloatDat */
Pfr->VectSize=(unsigned *)calloc(Pfr->NumFloatLines,sizeof(unsigned));
if(!Pfr->VectSize)Quit(MEMORY_ERR,"");

/* obtain float FloatDat sheet from char Sheet */
for(i=0;i<Pfr->NumFloatLines;i++){
Pfr->FloatDat[i]=GetFloatVectFromLine(Pfr->Sheet[Pfr->FloatLines[i]],&Pfr->VectSize[i],NULL);
if(Pfr->FloatDat[i]==NULL)Quit(PARF_ERR,Pfr->ParameterFile);
}

/* obtain string AlphaDat sheet from first string in Sheet line */
for(i=0;i<Pfr->NumAlphaLines;i++){
sscanf(Pfr->Sheet[Pfr->AlphaLines[i]],"%s",Pfr->AlphaDat[i]);
}


}


/*****************************************************************************
* FUNC: void ParamStorageID(PARFR *Pfr, unsigned LineNumber, unsigned *StorageIndex, char *Type)
*
* DESC: StorageIndex return the index in respective (FloatDat/AlphaDat)
*       storage sheet that corresponds to line LineNumber.
*
*       Type returns the respective data type: 'f'->float 'a'->alpha
*****************************************************************************/
void ParamStorageID(PARFR *Pfr, unsigned LineNumber, unsigned *StorageIndex, char *Type)
{

register unsigned i;
int p=-1;
char ParType=0;

//printf("LineNumber=%u\n",LineNumber);

/* scan FloatLines index array */
for(i=0;i<Pfr->NumFloatLines;i++){
if(Pfr->FloatLines[i]==LineNumber){p=i;ParType='f';break;}
}

/* if not found, scan AlphaLines index array */
if(p==-1){
for(i=0;i<Pfr->NumAlphaLines;i++){
if(Pfr->AlphaLines[i]==LineNumber){p=i;ParType='a';break;}
}

}

/* not found */
if(p==-1){fprintf(stderr,"No such line number in parameter file %s!\n",Pfr->ParameterFile);putch(BEL);exit(1);}

/* found */
*StorageIndex=p;
*Type=ParType;

}



/*****************************************************************************
* FUNC: unsigned long GetSheetNumLines(FILE *Fp, unsigned *MaxLength)
*
* DESC: Return number of lines of the text sheet in file pointed by Fp.
*       MaxLength return the line max length, including the '\0'.
******************************************************************************/
unsigned long GetSheetNumLines(FILE *Fp, unsigned *MaxLength)
{

  char line[MAXLINESIZE];
  register unsigned long Count;
  unsigned Length;

  *MaxLength=0;


  Count=0;
  while (fgets(line,MAXLINESIZE-2,Fp)){
  if((Length=strlen(line))>*MaxLength)*MaxLength=Length;
  Count++;
  }

  if(strlen(line)==1)Count--;/* to discount the last \n in file
                              for the case the user has typed it */

  *MaxLength= (*MaxLength)+1; /* to handle the '\0' added by fgets() */

  rewind(Fp);
  return Count;


}

/*****************************************************************************
* FUNC: char **CharMatCAlloc(unsigned long NRows,unsigned NCols)
*
* DESC: Allocate memory for a NRows x NCols  char matrix.
*****************************************************************************/
char **CharMatCAlloc(unsigned long NRows,unsigned NCols)
{
char  *x;
char  **Mx;
register unsigned  i;

x = (char  *)calloc((unsigned long)(NRows*NCols),sizeof(char));
if(!x)Quit(MEMORY_ERR,"");

Mx=(char  **)calloc((unsigned long)NRows,sizeof(char *));
if(!Mx)Quit(MEMORY_ERR,"");


for(i=0;i<NRows;i++){

Mx[i]=(char *)&x[i*NCols];

}

return Mx;
}


/*****************************************************************************
* FUNC: unsigned NumberOfFloatsInLine(char *Line)
*
* DESC: Return the number of floats in Line, starting from the beginning.
*       Upon finding a non-readable character, the function ends the count.
*****************************************************************************/
unsigned NumberOfFloatsInLine(char *Line)
{
static char Str[MAXLINESIZE];
static char Buf[MAXLINESIZE];

char *NonReadable;
double Val;
unsigned Count=0;

#if(0)
puts(Line);
#endif

while(1){
sprintf(Buf,"%s",Line);
if(Buf[0]=='\t'){Line++;continue;} /* found a tab */
if(Buf[0]==' '){Line++;continue;}  /* found a space */
if(!Buf[0])break; /* found a '\0' */
if(Buf[0]=='\n')break; /* found a '\n' */
sscanf(Buf,"%s",Str);
Val= strtod(Str, &NonReadable);

#if(0)
if(NonReadable[0]&&NonReadable[0]!='i')break; /* Check non-readable except 'i'-> to handle complex numbers */
#else
if(NonReadable[0])break; /* Check non-readable */
#endif

#if(0)
fprintf(stdout,"Val[%u]=%8.6lg%c\n",Count,Val,NonReadable[0]);
#endif

Line+=strlen(Str)+1;
Count++;
}

#if(0)
fprintf(stdout,"Line has %u floats.\n",Count);
#endif

return Count;
}

/*****************************************************************************
* FUNC: float *GetFloatVectFromLine(char *Line, unsigned *VectSize, float *Vector)
*
* DESC:  Convert the floats in Line into a float vector Vector, starting from the
*        beginning.
*
*        Upon finding a non-readable character, the function ends the count.
*        VectSize returns the size of Vector. If Vector==NULL, then memory
*        is allocated for the float vector Vector got from Line.
*****************************************************************************/
float *GetFloatVectFromLine(char *Line, unsigned *VectSize, float *Vector)
{
static char Str[MAXLINESIZE];
static char Buf[MAXLINESIZE];

char *NonReadable;
double Val;
unsigned Count=0;

*VectSize=NumberOfFloatsInLine(Line);
if(*VectSize==0)return NULL;

if(Vector==NULL){
Vector=(float *)calloc(*VectSize,sizeof(float));
if(!Vector)Quit(MEMORY_ERR,"");
}


while(1){
sprintf(Buf,"%s",Line);
if(Buf[0]=='\t'){Line++;continue;} /* found a tab */
if(Buf[0]==' '){Line++;continue;}  /* found a space */
if(!Buf[0])break; /* found a '\0' */
if(Buf[0]=='\n')break; /* found a '\n' */
sscanf(Buf,"%s",Str);
Val= strtod(Str, &NonReadable);

#if(0)
if(NonReadable[0]&&NonReadable[0]!='i')break; /* Check non-readable except 'i'-> to handle complex numbers */
#else
if(NonReadable[0])break; /* Check non-readable */
#endif

Vector[Count]=AULIM(Val); /* check float absolute upper limit */
if(Vector[Count]!=(float)Val){putch(BEL);fprintf(stderr,"Exceeded float maximum absolute value reading parameter from disk!");exit(1);}
Line+=strlen(Str)+1;
Count++;
}

return Vector;
}



/***************************************************************************
* FUNC: void Help(void)
*
* DESC: Help() shows the help menu.
****************************************************************************/
void Help(void)
{

system("cls");
textcolor(YELLOW);
gotoxy(22,2);cputs("Hebbian Neural Network Trainer V1.6.d");
textcolor(LIGHTBLUE);
gotoxy(18,4);cputs("By Fernando & Cristina de Castro - October 2001");
gotoxy(1,7);textcolor(LIGHTCYAN);cputs("Use: HNNTr ParameterFile\n");
textcolor(LIGHTGRAY);
exit(1);
}


/*****************************************************************************
* FUNC: void Automatic_ParamFileReader(PARFR *Pfr,char *ParameterFile, char AllocFlag)
*
* DESC: Reads automatically a parameter file.If AllocFlag!=0, memory is allocated.
*****************************************************************************/
void Automatic_ParamFileReader(PARFR *Pfr,char *ParameterFile, char AllocFlag)
{

#define SIZE_GUARD 6 /* a guard size for Pfr->Sheet line length to avoid eventual diferent float read from ParameterFile */

register unsigned i,j,k;

Pfr->AllocFlag=AllocFlag;
Pfr->ParameterFile=ParameterFile;


Pfr->NumFloatLines=0;
Pfr->NumAlphaLines=0;


/* Alloc memory for disk I/O speed-up buffers */
if(AllocFlag){
Pfr->BufferI=(char *)calloc(BUFSIZE,sizeof(char));
if(!Pfr->BufferI)Quit(MEMORY_ERR,"");
}


/* Open parameter file */
if((Pfr->IFp=fopen(Pfr->ParameterFile,"rt"))==NULL)Quit(READ_OPEN_ERR,Pfr->ParameterFile);
if (setvbuf(Pfr->IFp, Pfr->BufferI, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,Pfr->ParameterFile);

/* Get the total number of lines and the max line length */
Pfr->NumLines=GetSheetNumLines(Pfr->IFp,&Pfr->LineMaxLength);


/* Alloc memory for Sheet */
if(AllocFlag){
Pfr->Sheet=CharMatCAlloc(Pfr->NumLines,Pfr->LineMaxLength+SIZE_GUARD);
}

/* Read sheet and store it into Sheet */
for(i=0;i<Pfr->NumLines;i++)fgets(Pfr->Sheet[i],Pfr->LineMaxLength,Pfr->IFp);

/* close parameter file */
fclose(Pfr->IFp);


/* get NumFloatLines and NumAlphaLines from char Sheet */
for(i=0;i<Pfr->NumLines;i++){

if(NumberOfFloatsInLine(Pfr->Sheet[i])==0)Pfr->NumAlphaLines++;
else Pfr->NumFloatLines++;

}

/* check */
if(Pfr->NumLines!=(Pfr->NumFloatLines+Pfr->NumAlphaLines))Quit(PARF_ERR,Pfr->ParameterFile);


/* Alloc memory for FloatLines and AlphaLines ID vectors */
if(AllocFlag){
Pfr->FloatLines=(unsigned *)calloc(Pfr->NumFloatLines,sizeof(unsigned));
if(!Pfr->FloatLines)Quit(MEMORY_ERR,"");

Pfr->AlphaLines=(unsigned *)calloc(Pfr->NumAlphaLines,sizeof(unsigned));
if(!Pfr->AlphaLines)Quit(MEMORY_ERR,"");
}

/* Identify FloatLines and AlphaLines indexes from char Sheet */
j=k=0;
for(i=0;i<Pfr->NumLines;i++){

if(NumberOfFloatsInLine(Pfr->Sheet[i])==0)Pfr->AlphaLines[j++]=i;
else Pfr->FloatLines[k++]=i;

}

/* Alloc memory for float FloatDat row pointers */
if(AllocFlag){
Pfr->FloatDat=(float **)calloc(Pfr->NumFloatLines,sizeof(float *));
if(!Pfr->FloatDat)Quit(MEMORY_ERR,"");

/* Alloc memory for alphanumeric data */
Pfr->AlphaDat=CharMatCAlloc(Pfr->NumAlphaLines,MAXARGLEN);

/* Alloc memory for the array that contains the size of each row of FloatDat */
Pfr->VectSize=(unsigned *)calloc(Pfr->NumFloatLines,sizeof(unsigned));
if(!Pfr->VectSize)Quit(MEMORY_ERR,"");
}

/* obtain float FloatDat sheet from char Sheet */
for(i=0;i<Pfr->NumFloatLines;i++){

if(AllocFlag)Pfr->FloatDat[i]=GetFloatVectFromLine(Pfr->Sheet[Pfr->FloatLines[i]],&Pfr->VectSize[i],NULL);
else Pfr->FloatDat[i]=GetFloatVectFromLine(Pfr->Sheet[Pfr->FloatLines[i]],&Pfr->VectSize[i],Pfr->FloatDat[i]);

if(Pfr->FloatDat[i]==NULL)Quit(PARF_ERR,Pfr->ParameterFile);
}

/* obtain string AlphaDat sheet from first string in Sheet line */
for(i=0;i<Pfr->NumAlphaLines;i++){
sscanf(Pfr->Sheet[Pfr->AlphaLines[i]],"%s",Pfr->AlphaDat[i]);
}


}


